# coding=utf-8
from __future__ import absolute_import, print_function

import contextlib
import copy

from suanpan import error, g
from suanpan.arguments.auto import AutoArg
from suanpan.components import Arguments, Component
from suanpan.log import logger
from suanpan.stream.objects import Context


class Handler(Component):
    def run(self, stream, message, *arg, **kwargs):
        self.callBeforeInitHooks()
        context = self.init(stream, message)
        self.callAfterInitHooks(context)

        self.callBeforeCallHooks(context)
        if not self.runFunc:
            raise error.ComponentHandlerNotSet()
        if not callable(self.runFunc):
            raise error.ComponentHandlerNotCallable(self.name)
        results = self.runFunc(stream, context, *arg, **kwargs)
        self.callAfterCallHooks(context)

        self.callBeforeSaveHooks(context)
        outputs = self.save(results, message=message)
        self.callAfterSaveHooks(context)

        self.callBeforeCleanHooks(context)
        self.clean()
        self.callAfterCleanHooks()

        return outputs

    def beforeInit(self):
        logger.debug(f"Handler {self.name} starting...")

    def init(self, stream, message):  # pylint: disable=unused-argument
        self.argsDict = self.getArgsDict(message)
        context = self._getContext(message)
        args = self.loadComponentArguments(self.argsDict)
        context.update(args=args)
        return context

    @contextlib.contextmanager
    def context(self, message):
        yield Context.froms(message=message)

    def getArgsDict(self, message):
        return message or {}

    def load(self, args, argsDict=None):
        argsDict = argsDict or self.argsDict
        logger.debug("Loading Arguments:")
        for key, value in argsDict.items():
            logger.debug(f"{key}: {value}")
        return Context.froms(argsDict)

    def loadComponentArguments(self, argsDict=None):
        argsDict = argsDict or self.argsDict
        logger.debug("Loading Component Arguments:")
        logger.debug(f"Inputs: {argsDict}")
        return Context.froms(argsDict)

    def save(self, results, args=None, message=None):  # pylint: disable=unused-argument
        logger.debug("Saving...")
        logger.debug(f"Outputs: {results}")
        return results

    def afterSave(self, context):  # pylint: disable=unused-argument
        logger.debug(f"Handler {self.name} done.")


class CallHandler(Handler):
    def init(self, stream, message):
        context = super().init(stream, message)
        context.args.update(stream.args)
        return context

    def load(self, args, argsDict=None):
        logger.debug("Loading Arguments:")
        argsDict = argsDict or self.argsDict
        argsDict = {
            f"{arg.key}": argsDict.get(f"in{i + 1}")
            for i, arg in enumerate(args)
            if argsDict.get(f"in{i + 1}") is not None
        }
        for arg in args:
            if isinstance(arg, AutoArg):
                arg.setBackend(argtype="inputs")
        args = self.loadFormatArguments(args, argsDict)
        return Arguments.froms(*args)

    def loadFormatArguments(self, arguments, argsDict):
        arguments = super().loadFormatArguments(arguments, argsDict)
        for argument in arguments:
            argument.clean()
        return arguments

    def loadComponentArguments(self, argsDict=None):
        logger.debug("Loading Component Arguments:")
        argsDict = argsDict or self.argsDict
        arguments = copy.deepcopy(self.getArguments(exclude="outputs"))
        argsDict = {
            f"{arg.key}": argsDict.get(f"in{i + 1}")
            for i, arg in enumerate(arguments)
            if argsDict.get(f"in{i + 1}") is not None
        }
        argsDict = argsDict or self.argsDict
        args = self.loadFormatArguments(arguments, argsDict)
        return Arguments.froms(*args)

    def save(self, results, args=None, message=None):
        logger.debug("Saving...")
        if args:
            for arg in args:
                if isinstance(arg, AutoArg):
                    arg.setBackend(argtype="outputs")
        else:
            args = copy.deepcopy(self.getArguments(include="outputs"))
        shortRequestID = self.shortenRequestID(message["id"])
        outputArgsDict = {
            arg.key: arg.getOutputTmpValue(
                "studio",
                g.userId,
                "tmp",
                g.appId,
                shortRequestID,
                g.nodeId,
                f"out{i + 1}",
            )
            for i, arg in enumerate(args)
        }
        args = self.loadCleanArguments(args, outputArgsDict)
        return self.saveOutputs(args, results)

    def saveOutputs(self, args, results):
        if results is not None:
            outputs = super().saveOutputs(args, results)
            outputs = self.formatAsOuts(args, outputs)
            outputs = self.stringifyOuts(outputs)
            return outputs
        return None

    def saveOneOutput(self, arg, results):
        result = super().saveOneOutput(arg, results)
        arg.clean()
        return result

    def saveArguments(self, arguments, results):
        result = super().saveArguments(arguments, results)
        for argument in arguments:
            argument.clean()
        return result

    def formatAsOuts(self, args, results):
        return {
            f"out{i + 1}": self.getArgumentValueFromDict(results, arg)
            for i, arg in enumerate(args)
        }

    def stringifyOuts(self, outs):
        return {k: str(v) for k, v in outs.items() if v is not None}

    def shortenRequestID(self, requestID):
        return requestID.replace("-", "")


class TriggerHandler(CallHandler):
    pass


class SIOHandler(Handler):
    def run(self, stream, message, sid, *arg, **kwargs):
        self.callBeforeInitHooks()
        context = self.init(stream, message, sid)
        self.callAfterInitHooks(context)

        self.callBeforeCallHooks(context)
        if not self.runFunc:
            raise error.ComponentHandlerNotSet()
        if not callable(self.runFunc):
            raise error.ComponentHandlerNotCallable(self.name)
        results = self.runFunc(stream, context, *arg, **kwargs)
        self.callAfterCallHooks(context)

        self.callBeforeSaveHooks(context)
        outputs = self.save(results, message=message)
        self.callAfterSaveHooks(context)

        self.callBeforeCleanHooks(context)
        self.clean()
        self.callAfterCleanHooks()

        return outputs

    def getArgsDict(self, message):
        # front-end input is not guaranteed to be in dictionary format
        return message if isinstance(message, dict) else {'message': message}

    def init(self, stream, message, sid):  # pylint: disable=unused-argument
        self.argsDict = self.getArgsDict(message)
        context = self._getContext(message)
        args = self.loadComponentArguments(self.argsDict)
        context.update(args=args, sid=sid)
        return context


class SIOStreamHandler(Handler):
    def run(self, stream, message, sid, iostream, *arg, **kwargs):
        self.callBeforeInitHooks()
        context = self.init(stream, message, sid, iostream)
        self.callAfterInitHooks(context)

        self.callBeforeCallHooks(context)
        if not self.runFunc:
            raise error.ComponentHandlerNotSet()
        if not callable(self.runFunc):
            raise error.ComponentHandlerNotCallable(self.name)
        results = self.runFunc(stream, context, *arg, **kwargs)
        self.callAfterCallHooks(context)

        self.callBeforeSaveHooks(context)
        outputs = self.save(results, message=message)
        self.callAfterSaveHooks(context)

        self.callBeforeCleanHooks(context)
        self.clean()
        self.callAfterCleanHooks()

        return outputs

    def init(self, stream, message, sid, iostream):  # pylint: disable=unused-argument
        self.argsDict = self.getArgsDict(message)
        context = self._getContext(message)
        args = self.loadComponentArguments(self.argsDict)
        context.update(args=args, sid=sid, stream=iostream)
        return context



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan.mstorage import vars


class Variables(vars.Variables):
    VARS_EXPIRE = 60

    def __init__(self, context, mstorage):
        super().__init__(mstorage)
        self.context = context

    def _getFullKey(self, key):
        return f"{self.context.message['id']}.{key}"

    def _getOriginKey(self, fullKey):
        prefixLen = len(self.context.message["id"]) + 1
        return fullKey[prefixLen:]

    def _getRegisterKey(self):
        return f"{self.context.message['id']}.vars"



import abc
import typing as t
from suanpan.objects import HasName

if t.TYPE_CHECKING:
    from suanpan.stypes.trigger import TriggerT
    from suanpan.stypes.argument import ArgumentT


class AppT(HasName):
    args: t.Annotated[t.Any, 'arguments of app']

    @property
    @abc.abstractmethod
    def trigger(self) -> 'TriggerT':
        """Get a trigger instance that can run task at periodic intervals."""
        ...

    @abc.abstractmethod
    def use(self, handlerClass):
        """Decorate a class to run.

        Arguments:
            handlerClass: The handler class to add.

        Example:
            >>> @app.use
            >>> class CustomCode(MessageHandler):
            ...     def run(self, context):
            ...     print('run ...')
        """
        ...

    @abc.abstractmethod
    def input(self, argument: ArgumentT):
        """Decorate a function to add an input argument.

        Arguments:
            argument: The input argument to add.

        Example:
            >>> @app.input(String(key="formula", default=""))
            >>> def SPFormula(context):
            ...     args = context.args
        """
        ...

    @abc.abstractmethod
    def output(self, argument):
        """Decorate a function to add an output argument.

        Arguments:
            argument: The output argument to add.

        Example:
            >>> @app.output(String(key="outputTable", required=True))
            >>> def SPHiveSql(context):
            ...     args = context.args
        """
        ...

    @abc.abstractmethod
    def param(self, argument):
        """Decorate a function to add an param argument.

        Arguments:
            argument: The param argument to add.

        Example:
            >>> @app.param(String(key="sql", required=True))
            >>> def SPHiveSql(context):
            ...     args = context.args
        """
        ...

    @abc.abstractmethod
    def column(self, argument):
        """alias name of :meth:`param`"""
        ...

    @abc.abstractmethod
    def beforeInit(self, hook):
        """run the `hook` function before call init function"""
        ...

    @abc.abstractmethod
    def afterInit(self, hook):
        """run the `hook` function after call init function"""
        ...

    @abc.abstractmethod
    def beforeCall(self, hook):
        """run the `hook` function before call stream process function"""
        ...

    @abc.abstractmethod
    def afterCall(self, hook):
        """run the `hook` function after call stream process function"""
        ...

    @abc.abstractmethod
    def beforeExit(self, hook):
        """run the `hook` function before stream process function"""
        ...

    @abc.abstractmethod
    def load(self, args, argsDict=None):
        """load `args` and `argsDict` into a new Arguments

        Returns:
            Arguments: The return value. True for success, False otherwise.
        """
        ...

    @abc.abstractmethod
    def send(self, results, queue=None, message=None, args=None):
        """send `message` to message queue"""
        ...

    @property
    @abc.abstractmethod
    def vars(self):
        """the variables of app"""
        ...

    @property
    @abc.abstractmethod
    def modules(self):
        """the modules of app"""
        ...



# coding=utf-8
from __future__ import absolute_import, print_function

import sys
import ctypes
import socket
import struct
import random
import gevent
import requests
from suanpan import g
from suanpan.log import logger


def _get_open_ports():
    """
        This function will return a list of ports (TCP/UDP) that the current
        machine is listening on. It's basically a replacement for parsing
        netstat output but also serves as a good example for using the
        IP Helper API:
        https://docs.microsoft.com/en-us/windows/win32/api/iphlpapi/nf-iphlpapi-gettcptable
     """
    if sys.platform != "win32":
        raise Exception("only support win32")

    port_list = []

    DWORD = ctypes.c_ulong
    NO_ERROR = 0
    NULL = ""
    bOrder = 0

    # define some MIB constants used to identify the state of a TCP port
    MIB_TCP_STATE_CLOSED = 1
    MIB_TCP_STATE_LISTEN = 2
    MIB_TCP_STATE_SYN_SENT = 3
    MIB_TCP_STATE_SYN_RCVD = 4
    MIB_TCP_STATE_ESTAB = 5
    MIB_TCP_STATE_FIN_WAIT1 = 6
    MIB_TCP_STATE_FIN_WAIT2 = 7
    MIB_TCP_STATE_CLOSE_WAIT = 8
    MIB_TCP_STATE_CLOSING = 9
    MIB_TCP_STATE_LAST_ACK = 10
    MIB_TCP_STATE_TIME_WAIT = 11
    MIB_TCP_STATE_DELETE_TCB = 12

    ANY_SIZE = 1

    # defing our MIB row structures
    class MIB_TCPROW(ctypes.Structure):
        _fields_ = [('dwState', DWORD),
                    ('dwLocalAddr', DWORD),
                    ('dwLocalPort', DWORD),
                    ('dwRemoteAddr', DWORD),
                    ('dwRemotePort', DWORD)]

    class MIB_UDPROW(ctypes.Structure):
        _fields_ = [('dwLocalAddr', DWORD),
                    ('dwLocalPort', DWORD)]

    dwSize = DWORD(0)

    # call once to get dwSize
    ctypes.windll.iphlpapi.GetTcpTable(NULL, ctypes.byref(dwSize), bOrder)

    # ANY_SIZE is used out of convention (to be like MS docs); even setting this
    # to dwSize will likely be much larger than actually necessary but much
    # more efficient that just declaring ANY_SIZE = 65500.
    # (in C we would use malloc to allocate memory for the *table pointer and
    #  then have ANY_SIZE set to 1 in the structure definition)

    ANY_SIZE = dwSize.value

    class MIB_TCPTABLE(ctypes.Structure):
        _fields_ = [('dwNumEntries', DWORD),
                    ('table', MIB_TCPROW * ANY_SIZE)]

    tcpTable = MIB_TCPTABLE()
    tcpTable.dwNumEntries = 0  # define as 0 for our loops sake

    # now make the call to GetTcpTable to get the data
    if (ctypes.windll.iphlpapi.GetTcpTable(ctypes.byref(tcpTable),
                                           ctypes.byref(dwSize), bOrder) == NO_ERROR):

        maxNum = tcpTable.dwNumEntries
        placeHolder = 0

        # loop through every connection
        while placeHolder < maxNum:

            item = tcpTable.table[placeHolder]
            placeHolder += 1

            # format the data we need (there is more data if it is useful -
            #    see structure definition)
            lPort = item.dwLocalPort
            lPort = socket.ntohs(lPort)
            lAddr = item.dwLocalAddr
            lAddr = socket.inet_ntoa(struct.pack('L', lAddr))
            portState = item.dwState

            # only record TCP ports where we're listening on our external
            #    (or all) connections
            if str(lAddr) != "127.0.0.1" and portState == MIB_TCP_STATE_LISTEN:
                port_list.append({"port": lPort, "type": "TCP"})

    else:
        print("Error occurred when trying to get TCP Table")

    dwSize = DWORD(0)

    # call once to get dwSize
    ctypes.windll.iphlpapi.GetUdpTable(NULL, ctypes.byref(dwSize), bOrder)

    ANY_SIZE = dwSize.value  # again, used out of convention

    #                            (see notes in TCP section)

    class MIB_UDPTABLE(ctypes.Structure):
        _fields_ = [('dwNumEntries', DWORD),
                    ('table', MIB_UDPROW * ANY_SIZE)]

    udpTable = MIB_UDPTABLE()
    udpTable.dwNumEntries = 0  # define as 0 for our loops sake

    # now make the call to GetUdpTable to get the data
    if (ctypes.windll.iphlpapi.GetUdpTable(ctypes.byref(udpTable),
                                           ctypes.byref(dwSize), bOrder) == NO_ERROR):

        maxNum = udpTable.dwNumEntries
        placeHolder = 0
        while placeHolder < maxNum:

            item = udpTable.table[placeHolder]
            placeHolder += 1
            lPort = item.dwLocalPort

            lPort = socket.ntohs(lPort)
            lAddr = item.dwLocalAddr

            lAddr = socket.inet_ntoa(struct.pack('L', lAddr))

            # only record UDP ports where we're listening on our external
            #    (or all) connections
            if str(lAddr) != "127.0.0.1":
                port_list.append({"port": lPort, "type": "UDP"})
    else:
        print("Error occurred when trying to get UDP Table")

    return port_list


class PortHelper(object):
    def __init__(self, port_min, port_max):
        self.port_min = port_min
        self.port_max = port_max

    def get_free_port(self):
        op = set(range(self.port_min, self.port_max))
        open_ports = _get_open_ports()
        for port in open_ports:
            if self.port_min <= port['port'] <= self.port_max:
                op.discard(port['port'])

        if not op:
            raise Exception("not enough free port")

        return random.sample(op, 1)[0]


helper = PortHelper(g.portStart, g.portEnd)


def need_free_port():
    if g.currentOS == "windows":
        return True

    return False


def get_free_port():
    if need_free_port():
        return helper.get_free_port()

    raise Exception("not required for non-Windows systems")


def register_server(node_port, real_port):
    if not need_free_port():
        return

    protocol = "https" if g.hostTls else "http"
    url = f"{protocol}://localhost:{g.backendPort}/app/service/register"
    param = {"appId": g.appId, "nodeId": g.nodeId, "userId": g.userId, "nodePort": node_port, "port": real_port}

    while True:
        try:
            res = requests.post(url, json=param)
            res.raise_for_status()
            ret = res.json()
            if not ret["success"]:
                raise Exception(ret["msg"])

            return
        except Exception as e:
            logger.warn(f"register server: {e}")
            gevent.sleep(10)

# if __name__ == "__main__":
#     helper = PortHelper(5000, 6000)
#     p = helper.get_free_port()
#     print(p)



# coding=utf-8
from __future__ import absolute_import, print_function

import inspect
from unittest import mock

from suanpan import error
from suanpan.log import logger
from suanpan.objects import HasName
from suanpan.app import MessageHandler


class BaseApp(HasName):
    def __call__(self, *args, **kwargs):
        raise NotImplementedError("Method not implemented!")

    def start(self, *args, **kwargs):  # pylint: disable=unused-argument
        logger.info(f"Suanpan component {self.name} start...")

    def use(self, handlerClass):
        if inspect.isclass(handlerClass) and issubclass(handlerClass, MessageHandler):
            handler = handlerClass()
        elif isinstance(handlerClass, MessageHandler):
            handler = handlerClass
        else:
            raise error.AppError(f"can't use {handlerClass}")

        for arg in handler.inputs:
            self.input(arg)
        for arg in handler.outputs:
            self.output(arg)
        for arg in handler.params:
            self.param(arg)

        handler.init_params()
        self.beforeInit(handler.beforeInit)
        self.afterInit(handler.afterInit)
        self.beforeCall(handler.beforeCall)
        self.afterCall(handler.afterCall)

        self(handler.run)
        handler.initialize()

    @property
    def trigger(self):
        raise NotImplementedError(f"{self.name} not support trigger")

    def input(self, argument):
        raise NotImplementedError("Method not implemented!")

    def output(self, argument):
        raise NotImplementedError("Method not implemented!")

    def param(self, argument):
        raise NotImplementedError("Method not implemented!")

    def column(self, argument):
        raise NotImplementedError("Method not implemented!")

    def beforeInit(self, hook):
        raise NotImplementedError("Method not implemented!")

    def afterInit(self, hook):
        raise NotImplementedError("Method not implemented!")

    def beforeCall(self, hook):
        raise NotImplementedError("Method not implemented!")

    def afterCall(self, hook):
        raise NotImplementedError("Method not implemented!")

    def beforeExit(self, hook):
        raise NotImplementedError("Method not implemented!")

    def load(self, *args, **kwargs):
        raise NotImplementedError(f"{self.name} not support load")

    def save(self, *args, **kwargs):
        raise NotImplementedError(f"{self.name} not support save")

    def send(self, *args, **kwargs):
        raise NotImplementedError(f"{self.name} not support send")

    @property
    def args(self):
        raise NotImplementedError(f"{self.name} not support args")

    @property
    def vars(self):
        raise NotImplementedError(f"{self.name} not support args")

    def title(self, title):  # pylint: disable=unused-argument
        return self

    @property
    def modules(self):
        return mock.MagicMock()



# coding=utf-8
from __future__ import absolute_import, division, print_function

import functools
import os
from threading import Thread

from lostc import collection as lcc

import minio.error
from minio import Minio
from minio.commonconfig import CopySource
from suanpan import api, asyncio, g
from suanpan import path as spath
from suanpan import runtime
from suanpan.log import logger
from suanpan.storage import base
from suanpan.utils import pbar as spbar


class Storage(base.Storage):
    def __init__(
        self,
        minioAccessKey=None,
        minioSecretKey=None,
        minioBucketName="suanpan",
        minioEndpoint="minio-serivce.default:9000",
        minioSecure=True,
        minioDelimiter="/",
        minioTempStore=base.Storage.DEFAULT_TEMP_DATA_STORE,
        minioGlobalStore=base.Storage.DEFAULT_GLOBAL_DATA_STORE,
        **kwargs,
    ):
        super().__init__(
            delimiter=minioDelimiter,
            tempStore=minioTempStore,
            globalStore=minioGlobalStore,
            **kwargs,
        )

        self.bucketName = minioBucketName
        self.bucket = minioBucketName
        self.endpoint, self.secure = self._analyzeEndpoint(minioEndpoint, minioSecure)
        self.refreshAccessKey(accessKey=minioAccessKey, secretKey=minioSecretKey)

    def _analyzeEndpoint(self, endpoint, secure=False):
        httpsPrefix = "https://"
        if endpoint.startswith(httpsPrefix):
            return endpoint[len(httpsPrefix) :], True

        httpPrefix = "http://"
        if endpoint.startswith(httpPrefix):
            return endpoint[len(httpPrefix) :], False

        return endpoint, secure

    def autoRefreshToken(self, func):
        @functools.wraps(func)
        def _dec(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except minio.error.S3Error as e:
                if e.code == "AccessDenied":
                    logger.warning("Minio access denied, refreshing access key.")
                    self.refreshAccessKey()
                    return func(*args, **kwargs)
                raise e

        return _dec

    def refreshAccessKey(self, accessKey=None, secretKey=None):
        if accessKey and secretKey:
            self.accessKey = accessKey
            self.secretKey = secretKey
        else:
            data = api.oss.getToken()
            self.accessKey = data["Credentials"]["AccessKeyId"]
            self.secretKey = data["Credentials"]["AccessKeySecret"]

        self.client = Minio(
            self.endpoint,
            access_key=self.accessKey,
            secret_key=self.secretKey,
            secure=self.secure,
        )
        return self.accessKey, self.secretKey

    @runtime.retry(stop_max_attempt_number=3)
    def download(
        self,
        key,
        path=None,
        progress=None,
        bucket=None,
        ignores=None,
        quiet=not g.debug,
    ):
        ignores = ignores or self.DEFAULT_IGNORE_KEYWORDS
        downloadFunction = (
            self.downloadFile
            if self.isFile(key, bucket=bucket)
            else self.downloadFolder
        )
        return downloadFunction(
            key, path, progress=progress, bucket=bucket, ignores=ignores, quiet=quiet
        )

    def downloadFolder(
        self,
        key,
        path=None,
        progress=None,
        bucket=None,
        delimiter=None,
        ignores=None,
        quiet=not g.debug,
    ):
        bucket = bucket or self.bucket
        ignores = ignores or self.DEFAULT_IGNORE_KEYWORDS
        delimiter = delimiter or self.delimiter
        path = os.path.abspath(path) if path else self.getPathInTempStore(key)
        url = self.storageUrl(key, bucket=bucket)

        if not quiet:
            logger.info(f"Downloading folder: {url} -> {path}")

        if self.basename(key) in ignores:
            with spbar.one(
                total=1, desc="Downloading", disable=quiet, progress=progress
            ) as pbar:
                pbar.update(1)
                pbar.set_description("Ignored")
                return path

        downloads = {
            file: self.localPathJoin(path, self.storageRelativePath(file, key))
            for _, _, files in self.walk(key, delimiter=delimiter, bucket=bucket)
            for file in files
        }

        # Download from minio
        _run = functools.partial(
            self.downloadFile, bucket=bucket, ignores=ignores, quiet=True
        )
        asyncio.starmap(
            _run, downloads.items(), pbar=quiet, desc="Downloading", progress=progress,
        )
        # Remove ignores
        self.removeIgnores(path, ignores=ignores)
        # Remove rest files and folders
        files = (
            self.localPathJoin(root, file)
            for root, _, files in os.walk(path)
            for file in files
        )
        restFiles = [file for file in files if file not in downloads.values()]
        asyncio.map(
            spath.remove, restFiles, pbar=quiet, desc="Removing Rest Files",
        )
        spath.removeEmptyFolders(path)
        if not quiet:
            logger.debug(f"Removed empty folders in: {path}")
        # End
        return path

    def downloadFile(
        self,
        key,
        path=None,
        progress=None,
        bucket=None,
        ignores=None,
        quiet=not g.debug,
    ):
        bucket = bucket or self.bucket
        ignores = ignores or self.DEFAULT_IGNORE_KEYWORDS
        path = os.path.abspath(path) if path else self.getPathInTempStore(key)
        url = self.storageUrl(key, bucket=bucket)
        fileSize = self.getStorageSize(key, bucket=bucket)

        if not quiet:
            logger.info(f"Downloading file: {url} -> {path}")

        with spbar.one(
            total=fileSize, desc="Downloading", disable=quiet, progress=progress
        ) as pbar:
            if self.basename(key) in ignores:
                pbar.update(fileSize)
                pbar.set_description("Ignored")
                return path

            objectMd5 = self.getStorageMd5(key, bucket=bucket)
            fileMd5 = self.getLocalMd5(path)
            if self.checkMd5(objectMd5, fileMd5, bucket=bucket):
                pbar.update(fileSize)
                pbar.set_description("Existed")
                return path

            spath.safeMkdirsForFile(path)
            self.autoRefreshToken(self.client.fget_object)(bucket, key, path)

            pbar.update(fileSize)
            pbar.set_description("Downloaded")

            return path

    @runtime.retry(stop_max_attempt_number=3)
    def upload(
        self,
        name,
        path=None,
        progress=None,
        bucket=None,
        ignores=None,
        quiet=not g.debug,
    ):
        bucket = bucket or self.bucket
        ignores = ignores or self.DEFAULT_IGNORE_KEYWORDS
        uploadFunction = self.uploadFolder if os.path.isdir(path) else self.uploadFile
        return uploadFunction(
            name, path, progress=progress, bucket=bucket, ignores=ignores, quiet=quiet
        )

    def uploadFolder(
        self,
        key,
        path=None,
        progress=None,
        bucket=None,
        ignores=None,
        quiet=not g.debug,
    ):
        bucket = bucket or self.bucket
        ignores = ignores or self.DEFAULT_IGNORE_KEYWORDS
        path = os.path.abspath(path) if path else self.getPathInTempStore(key)
        url = self.storageUrl(key, bucket=bucket)

        if not quiet:
            logger.debug(f"Uploading folder: {path} -> {url}")

        if self.basename(key) in ignores:
            with spbar.one(
                total=1, desc="Uploading", disable=quiet, progress=progress
            ) as pbar:
                pbar.update(1)
                pbar.set_description("Ignored")
                return path

        files = (
            os.path.join(root, file)
            for root, _, files in os.walk(path)
            for file in files
        )
        uploads = {
            file: self.storagePathJoin(key, self.localRelativePath(file, path))
            for file in files
        }

        if not uploads:
            if not quiet:
                logger.warning(f"Uploading empty folder: {path}")
        else:
            # Upload files to oss
            uploadItems = [
                (objectName, filePath) for filePath, objectName in uploads.items()
            ]
            _run = functools.partial(
                self.uploadFile, bucket=bucket, ignores=ignores, quiet=True
            )
            asyncio.starmap(_run, uploadItems, pbar="Uploading")

        # Remove rest files
        localFiles = set(self.localRelativePath(file, path) for file in uploads.keys())
        remoteFiles = set(
            self.storageRelativePath(file, key)
            for _, _, files in self.walk(key, bucket=bucket)
            for file in files
        )
        restFiles = [
            self.storagePathJoin(key, file) for file in remoteFiles - localFiles
        ]
        _run = functools.partial(self.remove, bucket=bucket, quiet=True)
        asyncio.map(
            _run, restFiles, pbar=quiet, desc="Removing Rest Files", progress=progress,
        )
        # End
        return path

    def uploadFile(
        self,
        key,
        path=None,
        progress=None,
        bucket=None,
        ignores=None,
        quiet=not g.debug,
    ):
        bucket = bucket or self.bucket
        ignores = ignores or self.DEFAULT_IGNORE_KEYWORDS
        path = os.path.abspath(path) if path else self.getPathInTempStore(key)
        url = self.storageUrl(key, bucket=bucket)
        fileSize = os.path.getsize(path)

        if not quiet:
            logger.info(f"Uploading file: {path} -> {url}")

        with spbar.one(
            total=fileSize, desc="Uploading", disable=quiet, progress=progress
        ) as pbar:
            if self.basename(key) in ignores:
                pbar.update(fileSize)
                pbar.set_description("Ignored")
                return path

            objectMd5 = self.getStorageMd5(key, bucket=bucket)
            fileMd5 = self.getLocalMd5(path)
            if self.checkMd5(objectMd5, fileMd5):
                pbar.update(fileSize)
                pbar.set_description("Existed")
                return path

            self.autoRefreshToken(self.client.fput_object)(
                bucket,
                key,
                path,
                progress=Progress(pbar),
                metadata={self.CONTENT_MD5: fileMd5},
            )

            pbar.set_description("Uploaded")
            return path

    @runtime.retry(stop_max_attempt_number=3)
    def copy(self, src, dist, progress=None, bucket=None, quiet=not g.debug):
        bucket = bucket or self.bucket
        copyFunction = (
            self.copyFile if self.isFile(src, bucket=bucket) else self.copyFolder
        )
        return copyFunction(src, dist, progress=progress, bucket=bucket, quiet=quiet)

    def copyFolder(
        self, src, dist, progress=None, bucket=None, delimiter=None, quiet=not g.debug,
    ):
        bucket = bucket or self.bucket
        delimiter = delimiter or self.delimiter
        src = self.completePath(src)
        dist = self.completePath(dist)
        srcurl = self.storageUrl(src, bucket=bucket)
        disturl = self.storageUrl(dist, bucket=bucket)

        if not quiet:
            logger.info(f"Copying folder: {srcurl} -> {disturl}")

        copyItems = [
            (file, file.replace(src, dist))
            for _, _, files in self.walk(src, delimiter=delimiter, bucket=bucket)
            for file in files
        ]
        _run = functools.partial(self.copyFile, bucket=bucket, quiet=True)
        asyncio.starmap(
            _run, copyItems, pbar=quiet, desc="Copying", progress=progress,
        )
        return dist

    def copyFile(self, src, dist, progress=None, bucket=None, quiet=not g.debug):
        bucket = bucket or self.bucket
        fileSize = self.getStorageSize(src, bucket=bucket)
        srcurl = self.storageUrl(src, bucket=bucket)
        disturl = self.storageUrl(dist, bucket=bucket)

        if not quiet:
            logger.info(f"Copying file: {srcurl} -> {disturl}")

        with spbar.one(
            total=fileSize, desc="Copying", disable=quiet, progress=progress
        ) as pbar:
            objectMd5 = self.getStorageMd5(src, bucket=bucket)
            distMd5 = self.getStorageMd5(dist, bucket=bucket)
            if self.checkMd5(objectMd5, distMd5):
                pbar.update(fileSize)
                pbar.set_description("Existed")
                return dist

            # sourcePath = self.delimiter + self.storagePathJoin(bucket, src)
            self.autoRefreshToken(self.client.copy_object)(bucket, dist, CopySource(bucket, src))
            pbar.update(fileSize)
            pbar.set_description("Copied")
            return dist

    @runtime.retry(stop_max_attempt_number=3)
    def remove(
        self, key, progress=None, delimiter=None, bucket=None, quiet=not g.debug
    ):
        delimiter = delimiter or self.delimiter
        bucket = bucket or self.bucket
        removeFunc = (
            self.removeFile if self.isFile(key, bucket=bucket) else self.removeFolder
        )
        return removeFunc(key, delimiter=delimiter, bucket=bucket, quiet=quiet)

    def removeFolder(
        self, key, progress=None, delimiter=None, bucket=None, quiet=not g.debug
    ):
        delimiter = delimiter or self.delimiter
        bucket = bucket or self.bucket
        key = self.completePath(key)

        if not quiet:
            logger.info(f"Removing folder: {key}")

        removes = [
            objectName
            for _, _, files in self.walk(key, bucket=bucket, delimiter=delimiter)
            for objectName in files
        ]
        _run = functools.partial(
            self.removeFile, delimiter=delimiter, bucket=bucket, quiet=True
        )
        asyncio.map(
            _run,
            removes,
            pbar=quiet or not removes,
            desc="Removing",
            progress=progress,
        )
        return key

    def removeFile(
        self, key, progress=None, delimiter=None, bucket=None, quiet=not g.debug
    ):  # pylint: disable=unused-argument
        delimiter = delimiter or self.delimiter
        bucket = bucket or self.bucket

        if not quiet:
            logger.info(f"Removing file: {key}")

        _remove = self.autoRefreshToken(self.client.remove_object)
        with spbar.one(
            total=1, disable=quiet, desc="Removing", progress=progress,
        ) as pbar:
            self.autoRefreshToken(self.client.remove_object)(bucket, key)
            pbar.update(1)
            pbar.set_description("Removed")
            return key

    def walk(self, key, delimiter=None, bucket=None):
        bucket = bucket or self.bucket
        delimiter = delimiter or self.delimiter
        root = self.completePath(key, delimiter=delimiter)
        objects = runtime.saferun(self._listAll, default=iter(()))(
            key, delimiter=delimiter, bucket=bucket
        )
        folders, files = lcc.divide(objects, lambda obj: obj.is_dir)
        yield root, self._getObjectNames(folders), self._getObjectNames(files)
        for folder in folders:
            yield from self.walk(folder.object_name, delimiter=delimiter, bucket=bucket)

    def _listAll(self, key, delimiter=None, bucket=None):
        bucket = bucket or self.bucket
        delimiter = delimiter or self.delimiter
        prefix = self.completePath(key, delimiter=delimiter)
        return self.autoRefreshToken(self.client.list_objects)(bucket, prefix=prefix)

    def listAll(self, key, delimiter=None, bucket=None):
        return (obj.object_name for obj in self._listAll(key, delimiter=delimiter, bucket=bucket))

    def listFolders(self, key, delimiter=None, bucket=None):
        return (obj.object_name for obj in self._listAll(key, delimiter=delimiter, bucket=bucket) if obj.is_dir)

    def listFiles(self, key, delimiter=None, bucket=None):
        return (obj.object_name for obj in self._listAll(key, delimiter=delimiter, bucket=bucket) if not obj.is_dir)

    def isFolder(self, key, bucket=None):
        return bool(next(self.listAll(key, bucket=bucket), None))

    def isFile(self, key, bucket=None):
        bucket = bucket or self.bucket
        try:
            self.autoRefreshToken(self.client.stat_object)(bucket, key)
            return True
        except minio.error.S3Error as e:
            if e.code == "NoSuchKey":
                return False
            raise e

    def getStorageMd5(self, key, bucket=None):
        bucket = bucket or self.bucket
        try:
            return self.autoRefreshToken(self.client.stat_object)(
                bucket, key
            ).metadata.get(self.CONTENT_MD5)
        except minio.error.S3Error as e:
            if e.code == "NoSuchKey":
                return None
            raise e

    def getStorageSize(self, key, bucket=None):
        bucket = bucket or self.bucket
        return self.autoRefreshToken(self.client.stat_object)(bucket, key).size

    def storageUrl(self, key, bucket=None):
        bucket = bucket or self.bucket
        return "minio:///" + self.storagePathJoin(bucket, key)

    def _getObjectNames(self, objects):
        return (
            [obj.object_name for obj in objects]
            if isinstance(objects, (tuple, list))
            else objects.object_name
        )


class Progress(Thread):
    def __init__(self, pbar, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pbar = pbar
        self.size = None
        self.key = None

    def set_meta(self, total_length, object_name):
        self.size = total_length
        self.key = object_name

    def update(self, n):
        self.pbar.update(n)
        self.pbar.set_description("Uploading")



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan.error import base


class ArgumentError(base.Error):
    MESSAGE = "Argument Error: {}"


class ArgumentRequiredError(ArgumentError):
    MESSAGE = "Argument Required Error: {}"


class ArgumentTypeError(ArgumentError):
    MESSAGE = "Argument Type Error: {}"



# coding=utf-8
from __future__ import absolute_import, print_function

import os

import suanpan
from suanpan import api
from suanpan.app import app
from suanpan.app.arguments import String
from suanpan.utils import json


@app.param(String(key="app", required=True))
@app.param(String(key="graph", default=os.path.join("tmp", "graph")))
def SPAppGraphUpload(context):
    args = context.args
    filepath = (
        args.graph
        if args.graph.endswith(".json")
        else os.path.join(args.graph, f"{args.app}.json")
    )
    api.app.updateAppGraph(args.app, json.load(filepath))


if __name__ == "__main__":
    suanpan.run(app)



import abc
import os
import typing as t
from suanpan import g


class StorageT(object):
    tempStore: t.Annotated[str, 'the folder to save tmp files']
    globalStore: t.Annotated[str, 'the folder to save global files']
    delimiter = t.Annotated[str, 'default is os.sep']

    @abc.abstractmethod
    def download(self, key, path=None, progress=None, bucketOrBucketName=None, ignores=None, quiet=not g.debug):
        """download a file from remote storage, oss or minio

        Arguments:
            key: the file `key` in storage
            path: save file to `path`
            progress: print a progress bar
            bucketOrBucketName: a bucket instance or bucket name
            ignores: ignore file names
            quiet: if False, logging download message
        """
        ...

    @abc.abstractmethod
    def upload(self, key, path=None, progress=None, bucketOrBucketName=None, ignores=None, quiet=not g.debug):
        """upload a file to remote storage, oss or minio

        Arguments:
            key: the file `key` in storage
            path: upload file from `path`
            progress: print a progress bar
            bucketOrBucketName: a bucket instance or bucket name
            ignores: ignore file names
            quiet: if False, logging upload message
        """
        ...

    @abc.abstractmethod
    def copy(self, name, dist, progress=None, bucketOrBucketName=None, quiet=not g.debug):
        """copy a file or folder from `name` to `dist` in remote storage, oss or minio

        Arguments:
            name: the source file or folder `name` in storage
            dist: the destination of copy
            progress: print a progress bar
            bucketOrBucketName: a bucket instance or bucket name
            quiet: if False, logging upload message
        """
        ...

    @abc.abstractmethod
    def remove(self, key, progress=None, delimiter=None, bucketOrBucketName=None, quiet=not g.debug):
        """remove a file or folder from in remote storage, oss or minio

        Arguments:
            key: the file `key` to remove
            progress: print a progress bar
            delimiter: delimiter of path
            bucketOrBucketName: a bucket instance or bucket name
            quiet: if False, logging upload message
        """
        ...

    @abc.abstractmethod
    def walk(self, key, delimiter=None, bucketOrBucketName=None):
        """Generate the file names in a directory tree by walking the tree.

        Arguments:
            key: the path `key` to walk
            delimiter: delimiter of path
            bucketOrBucketName: a bucket instance or bucket name
        """
        ...

    @abc.abstractmethod
    def listAll(self, key, delimiter=None, bucketOrBucketName=None):
        """Return a list containing the names of the entries in the directory given by path.

        Arguments:
            key: the path `key` to list
            delimiter: delimiter of path
            bucketOrBucketName: a bucket instance or bucket name
        """
        ...

    @abc.abstractmethod
    def listFolders(self, key, delimiter=None, bucketOrBucketName=None):
        """Return a list containing the names of the folders in the directory given by path.

        Arguments:
            key: the path `key` to list
            delimiter: delimiter of path
            bucketOrBucketName: a bucket instance or bucket name
        """
        ...

    @abc.abstractmethod
    def listFiles(self, key, delimiter=None, bucketOrBucketName=None):
        """Return a list containing the names of the files in the directory given by path.

        Arguments:
            key: the path `key` to list
            delimiter: delimiter of path
            bucketOrBucketName: a bucket instance or bucket name
        """
        ...

    @property
    def appStoreKey(self):
        return self.storagePathJoin("studio", g.userId, g.appId)

    @property
    def appDataStoreKey(self):
        return self.storagePathJoin("studio", g.userId, "share", g.appId)

    @property
    def appConfigsStoreKey(self):
        return self.storagePathJoin("studio", g.userId, "configs", g.appId)

    @property
    def appTmpStoreKey(self):
        return self.storagePathJoin("studio", g.userId, "tmp", g.appId)

    @property
    def appLogsStoreKey(self):
        return self.storagePathJoin("studio", g.userId, "logs", g.appId)

    @property
    def nodeStoreKey(self):
        return self.storagePathJoin("studio", g.userId, g.appId, g.nodeId)

    @property
    def nodeDataStoreKey(self):
        return self.storagePathJoin("studio", g.userId, "share", g.appId, g.nodeId)

    @property
    def nodeConfigsStoreKey(self):
        return self.storagePathJoin("studio", g.userId, "configs", g.appId, g.nodeId)

    @property
    def nodeLogsStoreKey(self):
        return self.storagePathJoin("studio", g.userId, "logs", g.appId, g.nodeId)

    @property
    def componentsStoreKey(self):
        return self.storagePathJoin("studio", g.userId, "component")

    def getKeyInAppStore(self, *paths):
        return self.storagePathJoin(self.appStoreKey, *paths)

    def getKeyInAppDataStore(self, *paths):
        return self.storagePathJoin(self.appDataStoreKey, *paths)

    def getKeyInAppConfigsStore(self, *paths):
        return self.storagePathJoin(self.appConfigsStoreKey, *paths)

    def getKeyInAppTmpStore(self, requestID, *paths):
        return self.storagePathJoin(self.appTmpStoreKey, requestID, *paths)

    def getKeyInAppLogsStore(self, *paths):
        return self.storagePathJoin(self.appLogsStoreKey, *paths)

    def getKeyInNodeStore(self, *paths):
        return self.storagePathJoin(self.nodeStoreKey, *paths)

    def getKeyInNodeDataStore(self, *paths):
        return self.storagePathJoin(self.nodeDataStoreKey, *paths)

    def getKeyInNodeConfigsStore(self, *paths):
        return self.storagePathJoin(self.nodeConfigsStoreKey, *paths)

    def getKeyInNodeTmpStore(self, requestID, *paths):
        return self.storagePathJoin(self.appTmpStoreKey, requestID, g.nodeId, *paths)

    def getKeyInNodeLogsStore(self, *paths):
        return self.storagePathJoin(self.nodeLogsStoreKey, *paths)

    def getKeyInComponentDataStore(self, componentId, *paths):
        return self.storagePathJoin(self.componentsStoreKey, componentId, *paths)

    def getPathInTempStore(self, *path):
        return self.localPathJoin(self.tempStore, *path)

    def getPathInAppStore(self, *paths):
        return self.getPathInTempStore(self.getKeyInAppStore(*paths))

    def getPathInAppDataStore(self, *paths):
        return self.getPathInTempStore(self.getKeyInAppDataStore(*paths))

    def getPathInAppConfigsStore(self, *paths):
        return self.getPathInTempStore(self.getKeyInAppConfigsStore(*paths))

    def getPathInAppTmpStore(self, requestID, *paths):
        return self.getPathInTempStore(self.getKeyInAppTmpStore(requestID, *paths))

    def getPathInAppLogsStore(self, *paths):
        return self.getPathInTempStore(self.getKeyInAppLogsStore(*paths))

    def getPathInNodeDataStore(self, *paths):
        return self.getPathInTempStore(self.getKeyInNodeDataStore(*paths))

    def getPathInNodeConfigsStore(self, *paths):
        return self.getPathInTempStore(self.getKeyInNodeConfigsStore(*paths))

    def getPathInNodeTmpStore(self, requestID, *paths):
        return self.getPathInTempStore(self.getKeyInNodeLogsStore(requestID, *paths))

    def getPathInNodeLogsStore(self, *paths):
        return self.getPathInTempStore(self.getKeyInNodeLogsStore(*paths))

    def getPathInGlobalStore(self, *path):
        return self.localPathJoin(self.globalStore, *path)

    def getPathInGlobalDataStore(self, *paths):
        return self.getPathInGlobalStore("data", *paths)

    def getPathInGlobalConfigsStore(self, *paths):
        return self.getPathInGlobalStore("configs", *paths)

    def completePath(self, path, delimiter=None):
        delimiter = delimiter or self.delimiter
        return path if path.endswith(delimiter) else path + delimiter

    def toLocalPath(self, path, delimiter=None):
        delimiter = delimiter or self.delimiter
        return str(path).replace(delimiter, os.sep)

    def toStoragePath(self, path, delimiter=None):
        delimiter = delimiter or self.delimiter
        return str(path).replace(os.sep, delimiter)

    def localPathJoin(self, *paths):
        path = os.path.join(*paths)
        return self.toLocalPath(path)

    def storagePathJoin(self, *paths):
        path = os.path.join(*[self.toLocalPath(path) for path in paths])
        return self.toStoragePath(path)

    def localRelativePath(self, path, base):
        return self._relativePath(path, base, delimiter=os.sep)

    def storageRelativePath(self, path, base, delimiter=None):
        delimiter = delimiter or self.delimiter
        return self._relativePath(path, base, delimiter=delimiter)

    def _relativePath(self, path, base, delimiter):
        base = base if base.endswith(delimiter) else base + delimiter
        return path[len(base):] if path.startswith(base) else path



import abc
import typing as t
from suanpan.components import Arguments
from suanpan.stream.objects import Context
from suanpan.interfaces import HasArguments

if t.TYPE_CHECKING:
    from suanpan.stypes import ArgumentT


class MessageHandler(HasArguments):
    def __init__(self):
        self._inputs = []
        self._outputs = []
        self._params = []
        self._context = Context()

    def init_params(self):
        args_dict = self.getArgsDict()
        _params = self.loadFormatArguments(self.params, args_dict)
        _args = Arguments.froms(*_params)
        self._context = Context.froms(args=_args)

    @property
    def context(self):
        return self._context

    @property
    def inputs(self):
        return self._inputs

    @property
    def outputs(self):
        return self._outputs

    @property
    def params(self):
        return self._params

    def add_input(self, argument: 'ArgumentT'):
        self._inputs.append(argument)

    def add_output(self, argument: 'ArgumentT'):
        self._outputs.append(argument)

    def add_param(self, argument: 'ArgumentT'):
        self._params.append(argument)

    def beforeInit(self):
        ...

    def afterInit(self, context):
        ...

    def beforeCall(self, context):
        ...

    def afterCall(self, context):
        ...

    def initialize(self):
        """initialize before message loop"""
        ...

    @abc.abstractmethod
    def run(self, context):
        """run message loop"""
        ...

    def getArguments(self, *args, **kwargs):
        ...



class DataWarehouse(object):
    def __init__(self, *args, **kwargs):
        pass

    def readTable(self, table, **kwargs):
        raise NotImplementedError("Must install python with Table support")

    def writeTable(self, table, data, **kwargs):
        raise NotImplementedError("Must install python with Table support")



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan import error, g
from suanpan.api import requests


def getAffinityUrl(path, tls=g.hostTls):
    if g.affinity:
        return g.affinity + path

    if g.host:
        protocol = "https" if tls else "http"
        host = f"{g.host}:{g.backendPort}" if g.currentOS == "windows" else g.host
        return f"{protocol}://{host}{path}"

    raise error.ApiError("affinity and host not set")


def get(path, *args, **kwargs):
    return requests.get(getAffinityUrl(path), *args, **kwargs)


def post(path, *args, **kwargs):
    return requests.post(getAffinityUrl(path), *args, **kwargs)


def put(path, *args, **kwargs):
    return requests.put(getAffinityUrl(path), *args, **kwargs)


def delete(path, *args, **kwargs):
    return requests.delete(getAffinityUrl(path), *args, **kwargs)



# coding=utf-8
from __future__ import absolute_import, print_function

import typing
from suanpan import g
from suanpan.proxy import Proxy
from suanpan.app.message_handler import MessageHandler

if typing.TYPE_CHECKING:
    from suanpan.stypes import AppT


class App(Proxy):
    MAPPING = {
        "spark": "suanpan.app.spark.SparkApp",
        "docker": "suanpan.app.docker.DockerApp",
        "stream": "suanpan.app.stream.StreamApp",
    }

    @property
    def isSpark(self):
        return self.isType("spark")

    @property
    def isDocker(self):
        return self.isType("docker")

    @property
    def isStream(self):
        return self.isType("stream")

    def isType(self, appType):
        return getattr(self, self.TYPE_KEY, None) == appType


TYPE = g.appType
app: 'AppT' = App(type=TYPE)



# coding=utf-8
from __future__ import absolute_import, print_function

import os
from suanpan import path, runtime
from suanpan.arguments import Arg
from suanpan.components import Result
from suanpan.storage import storage
from suanpan.log import logger
from suanpan.utils import csv, excel, image, json, npy, text


class StorageArg(Arg):
    def getOutputTmpValue(self, *args):
        return storage.delimiter.join(args)


class File(StorageArg):
    FILENAME = "file"
    FILETYPE = None

    def __init__(self, key, **kwargs):
        fileName = kwargs.pop("name", self.FILENAME)
        fileType = kwargs.pop("type", self.FILETYPE)
        self.fileName = f"{fileName}.{fileType.lower()}" if fileType else fileName
        self.objectPrefix = None
        self.objectName = None
        self.filePath = None
        super().__init__(key, **kwargs)

    @property
    def isSet(self):
        return True

    def load(self, args):
        super().load(args)
        if self.value:
            self.objectPrefix = self.value
            self.objectName = storage.storagePathJoin(self.objectPrefix, self.fileName)
            self.filePath = storage.getPathInTempStore(self.objectName)
            self.value = self.filePath
        return self

    def transform(self, value):
        if self.filePath:
            _download = (
                storage.download if self.required else runtime.saferun(storage.download)
            )
            _download(self.objectName, self.filePath)
        return self.filePath

    def clean(self):
        if self.filePath:
            path.remove(self.filePath)
            logger.debug("clean folder from {} to {}".format(self.filePath, storage.tempStore))
            path.removeEmptyFoldersTo(os.path.dirname(self.filePath), storage.tempStore)
        return self

    def save(self, result):
        filePath = result.value
        storage.upload(self.objectName, filePath)
        if filePath != self.filePath:
            path.remove(filePath)
        self.logSaved(self.objectName)
        return self.objectPrefix


class Folder(StorageArg):
    def __init__(self, key, **kwargs):
        super().__init__(key, **kwargs)
        self.folderName = None
        self.folderPath = None

    @property
    def isSet(self):
        return True

    def load(self, args):
        super().load(args)
        if self.value:
            self.folderName = self.value
            self.folderPath = storage.getPathInTempStore(self.folderName)
            self.value = self.folderPath
        return self

    def transform(self, value):
        if self.folderPath:
            _download = (
                storage.download if self.required else runtime.saferun(storage.download)
            )
            _download(self.folderName, self.folderPath)
        return self.folderPath

    def clean(self):
        if self.folderPath:
            path.empty(self.folderPath)
        return self

    def save(self, result):
        folderPath = result.value
        storage.upload(self.folderName, folderPath)
        if folderPath != self.folderPath:
            path.remove(folderPath)
        self.logSaved(self.folderName)
        return self.folderName


class Data(File):
    FILENAME = "data"


class Json(Data):
    FILETYPE = "json"

    def transform(self, value):
        filePath = super().transform(value)
        if not filePath:
            return None
        _load = json.load if self.required else runtime.saferun(json.load)
        return _load(filePath)

    def save(self, result):
        path.mkdirs(self.filePath, parent=True)
        json.dump(result.value, self.filePath)
        return super().save(Result.froms(value=self.filePath))


class Csv(Data):
    FILETYPE = "csv"

    def cleanParams(self, params):
        params = super().cleanParams(params)
        return {k: v for k, v in params.items() if k not in ("table", "partition")}

    def transform(self, value):
        filePath = super().transform(value)
        if not filePath:
            return None
        _load = csv.load if self.required else runtime.saferun(csv.load)
        return _load(filePath)

    def save(self, result):
        path.mkdirs(self.filePath, parent=True)
        csv.dump(result.value, self.filePath)
        return super().save(Result.froms(value=self.filePath))


class Table(Csv):
    pass


class Excel(Data):
    FILETYPE = "xlsx"
    LEGACY_FILETYPE = "xls"

    def __init__(self, key, **kwargs):
        self.fileTypeSet = kwargs.get("type") is not None
        super().__init__(key, **kwargs)

    def load(self, args):
        super().load(args)
        if not self.objectName:
            self.value = self.filePath
            return self

        if not storage.isFile(self.objectName) and not self.fileTypeSet:
            fileType = self.LEGACY_FILETYPE
            self.fileName = "{}.{}".format(self.fileName.split(".")[0], fileType)
            self.objectName = storage.storagePathJoin(self.objectPrefix, self.fileName)
        self.filePath = storage.getPathInTempStore(self.objectName)
        path.safeMkdirsForFile(self.filePath)
        self.value = self.filePath
        return self

    def transform(self, value):
        value = super().transform(value)
        if self.filePath:
            _load = excel.load if self.required else runtime.saferun(excel.load)
            value = _load(self.filePath)
        return value

    def save(self, result):
        path.mkdirs(self.filePath, parent=True)
        excel.dump(result.value, self.filePath)
        return super().save(Result.froms(value=self.filePath))


class Npy(Data):
    FILETYPE = "npy"

    def transform(self, value):
        filePath = super().transform(value)
        if not filePath:
            return None
        _load = npy.load if self.required else runtime.saferun(npy.load)
        return _load(filePath)

    def save(self, result):
        path.mkdirs(self.filePath, parent=True)
        npy.dump(result.value, self.filePath)
        return super().save(Result.froms(value=self.filePath))


class String(Data):
    FILETYPE = "txt"

    def transform(self, value):
        filePath = super().transform(value)
        if not filePath:
            return None
        _load = text.load if self.required else runtime.saferun(text.load)
        return _load(filePath)

    def save(self, result):
        path.mkdirs(self.filePath, parent=True)
        text.dump(result.value, self.filePath)
        return super().save(Result.froms(value=self.filePath))


class Visual(String):
    FILENAME = "part-00000"
    FILETYPE = ""


class Model(File):
    FILENAME = "model"


class H5Model(Model):
    FILETYPE = "h5"


class Checkpoint(Model):
    FILETYPE = "ckpt"


class JsonModel(Model):
    FILETYPE = "json"


class Image(File):
    FILENAME = "image"
    FILETYPE = "png"

    def transform(self, value):
        filePath = super().transform(value)
        if not filePath:
            return None
        _read = image.read if self.required else runtime.saferun(image.read)
        return _read(filePath)

    def save(self, result):
        image.save(self.filePath, result.value)
        return super().save(Result.froms(value=self.filePath))


class Screenshots(Image):
    FILENAME = "screenshots"



# coding=utf-8
from __future__ import absolute_import, print_function

from contextlib import contextmanager

from pyspark.sql import SparkSession  # pylint: disable=import-error
from suanpan import g
from suanpan.components import Component
from suanpan.log import logger
from suanpan.objects import Context, HasName


class SparkComponent(Component, HasName):
    def beforeInit(self):
        logger.logDebugInfo()
        logger.debug(f"SparkComponent {self.name} starting...")
        g.spark = (
            SparkSession.builder.appName(self.runFunc.__name__)
            .enableHiveSupport()
            .getOrCreate()
        )

    def initBase(self, args):
        logger.setLogger(self.name)

    def afterSave(self, context):  # pylint: disable=unused-argument
        logger.debug(f"SparkComponent {self.name} done.")

    def afterClean(self):
        g.spark.stop()



# coding=utf-8
from __future__ import absolute_import, print_function

import socketio
from suanpan.api import auth


def sio(*args, **kwargs):
    kwargs["headers"] = {**auth.defaultHeaders(), **kwargs.pop("headers", {})}
    client = socketio.Client()
    client.connect(*args, **kwargs)

    return client
